#!/usr/bin/env python

"""
Translate the AST of a Python program into a more interpretable representation.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import *
from micropython.data import *
from micropython.rsvp import *
import compiler.ast
from compiler.visitor import ASTVisitor

# Program visitors.

class Translation(ASTVisitor):

    "A translated module."

    supported_optimisations = ["constant_storage", "known_target", "self_access", "temp_storage", "load_operations", "unused_results"]

    attribute_load_instructions = (LoadAddress, LoadAddressContext, LoadAttr, LoadAttrIndex)
    attribute_store_instructions = (None, None, StoreAttr, StoreAttrIndex)
    name_load_instructions = (LoadName, LoadAddress)
    name_store_instructions = (StoreName, StoreAddress)

    def __init__(self, module, importer, optimisations=None):

        """
        Initialise the translation with an inspected 'module', the 'importer'
        and optional 'optimisations'. See the 'supported_optimisations'
        attribute of this class for permitted values.
        """

        ASTVisitor.__init__(self)
        self.visitor = self
        self.module = module

        # Global program dependencies.

        self.importer = importer
        self.objtable = self.importer.get_object_table()
        self.paramtable = self.importer.get_parameter_table()
        self.builtins = self.importer.modules.get("__builtins__")

        # Desired optimisations.

        self.optimisations = set(optimisations or [])

        # The current unit being translated.

        self.unit = None

        # The current "active" instruction.
        # As a rule, this will become the last instruction, but some
        # control-flow operations will flush the "active" instruction.

        self.active = None

        # The temporary storage used by the current assignment expression.

        self.expr_temp = None

        # Wiring within the code.

        self.labels = {}
        self.label_number = 0
        self.loop_labels = []
        self.exception_labels = []

        # The code itself. This is limited to the code for a particular block
        # being processed. Also retained is information about temporary values.

        self.code = None
        self.temp_position = 0

    def __repr__(self):
        return "Translation(%r)" % self.module

    def get_module_code(self):

        "Return the top-level module code."

        self.unit = self.module
        self.code = []
        self.temp_position = 0

        if self.module.module is not None:
            self.dispatch(self.module.module)

        return self.code

    def get_code(self, unit):

        "Return the code for the given 'unit'."

        self.unit = unit
        self.code = []
        self.temp_position = 0

        if unit.astnode is not None:
            self.dispatch(unit.astnode)

        return self.code

    # Name-related methods.

    def get_scope(self, name):

        "Return the scope for the given 'name'."

        if self.unit.has_key(name):
            return "local"
        elif self.module.has_key(name):
            return "global"
        else:
            return "builtins"

    def load_builtin(self, name, node):

        "Generate an instruction loading 'name' for the given 'node'."

        self.new_op(LoadAddress(self.get_builtin(name, node)))

    def get_builtin(self, name, node):

        """
        Return the built-in module definition for the given 'name', used by the
        given 'node'.
        """

        if self.builtins is not None:
            try:
                return self.builtins[name]
            except KeyError:
                raise TranslateError(self.module.full_name(), node, "No __builtins__ definition is available for name %r." % name)
        else:
            raise TranslateError(self.module.full_name(), node, "No __builtins__ module is available for name %r." % name)

    # Code feature methods.

    def new_label(self):

        "Return a new label object for use with set_label."

        number = self.label_number
        label = Label(number)
        self.labels[label] = label
        self.label_number += 1
        return label

    def set_label(self, label):

        """
        Set the location of 'label' to that within the entire image: the
        location within the code combined with location of the code unit.
        """

        label.location = len(self.code) + self.unit.code_location

    def get_loop_labels(self):
        return self.loop_labels[-1]

    def add_loop_labels(self, next_label, exit_label):
        self.loop_labels.append((next_label, exit_label))

    def drop_loop_labels(self):
        self.loop_labels.pop()

    def get_exception_labels(self):
        return self.exception_labels[-1]

    def add_exception_labels(self, handler_label, exit_label):
        self.exception_labels.append((handler_label, exit_label))

    def drop_exception_labels(self):
        self.exception_labels.pop()

    # Assignment expression values.

    def record_value(self):
        self.expr_temp = self._optimise_temp_storage()

    def discard_value(self):
        self.discard_temp(self.expr_temp)
        self.expr_temp = None

    def set_source(self):
        if self.active is not None:
            self.active.source = self.expr_temp

        # Optimise away constant storage if appropriate.

        self._optimise_constant_storage()

    # Temporary storage administration.

    def get_temp(self):

        """
        Add a temporary storage instruction for the current value and return a
        sequence of access instructions.
        """

        temp_position = self.reserve_temp(1)
        self.new_op(StoreTemp(temp_position))
        return LoadTemp(temp_position)

    def reserve_temp(self, n):
        temp_position = self.temp_position
        self.temp_position += n
        return temp_position

    def discard_temp(self, instruction):
        if isinstance(instruction, LoadTemp):
            self.temp_position -= 1

    # Code writing methods.

    def new_op(self, op):

        "Add 'op' to the generated code."

        # Optimise load operations employed by this instruction.

        self._optimise_load_operations(op)

        self.code.append(op)
        self.active = op

    def remove_op(self):

        "Remove the last instruction."

        op = self.code.pop()
        self.active = None

    def replace_op(self, op):

        "Replace the last added instruction with 'op'."

        self.remove_op()
        self.new_op(op)

    def last_op(self):

        "Return the last added instruction."

        try:
            return self.code[-1]
        except IndexError:
            return None

    # Optimisation tests.

    def _should_optimise_constant_storage(self):
        return "constant_storage" in self.optimisations

    def _should_optimise_known_target(self):
        return "known_target" in self.optimisations

    def _should_optimise_self_access(self):
        return "self_access" in self.optimisations

    def _should_optimise_temp_storage(self):
        return "temp_storage" in self.optimisations

    def _should_optimise_load_operations(self):
        return "load_operations" in self.optimisations

    def _should_optimise_unused_results(self):
        return "unused_results" in self.optimisations

    # Simple tests.

    def _is_constant_input(self, instruction):

        "Return whether 'instruction' provides a constant input."

        return isinstance(instruction, LoadAddress) and instruction.attr.assignments == 1 or \
            isinstance(instruction, LoadConst)

    def _is_constant_target(self, instruction):

        "Return whether 'instruction' provides a constant target."

        return isinstance(instruction, (StoreName, StoreAddress)) and \
            instruction.attr.assignments == 1

    def _is_simple_input(self, instruction):

        "Return whether 'instruction' provides a simple input."

        return isinstance(instruction, (LoadConst, LoadName, LoadTemp, LoadResult, LoadAddress, MakeObject))

    def _is_simple_input_user(self, instruction):

        "Return whether 'instruction' can use simple input from the current value."

        return isinstance(instruction, (
            StoreTemp, StoreFrame, StoreResult, StoreException, # as the value being stored
            LoadAddressContext, LoadAttr, LoadAttrIndex,        # as the object being referenced
            StoreAttr, StoreAttrIndex, StoreCallable            # as the object being referenced
            ))

    def _is_input(self, instruction):

        "Return whether 'instruction' provides an input."

        return isinstance(instruction, (LoadConst, LoadName, LoadTemp, LoadResult) + self.attribute_load_instructions)

    # Convenience tests.

    def _have_constant_input(self):

        "Return whether the active instruction provides a constant input."

        return self._is_constant_input(self.active)

    _have_known_target = _have_constant_input

    def _have_constant_target(self):

        "Return whether the active instruction provides a constant target."

        return self._is_constant_target(self.active)

    def _have_constant_source(self):

        "Return whether the active instruction has a constant source."

        return self._is_constant_input(self.active.source)

    def _have_simple_input(self):

        "Return whether the active instruction provides a simple input."

        return self._is_simple_input(self.active)

    def _have_input(self):

        "Return whether the active instruction provides an input."

        return self._is_input(self.active)

    def _have_self_input(self):

        "Return whether the active instruction is a reference to self."

        return isinstance(self.unit, Function) and \
            self.unit.is_method() and isinstance(self.active, LoadName) and \
            self.active.attr.name == "self"

    def _have_temp_compatible_access(self):

        """
        Indicate whether the active instruction can be used in place of access
        to a temporary variable retaining the result of the last instruction.
        """

        return isinstance(self.active, (LoadName, LoadTemp, LoadAddress, LoadConst, LoadResult))

    # Optimisation methods. See the supported_optimisations class attribute.

    def _optimise_constant_storage(self):

        """
        Where the last operation stores a constant into a target which is also
        constant, optimise away both operations.
        """

        if self._should_optimise_constant_storage() and \
            self._have_constant_target() and \
            self._have_constant_source():

            self.remove_op()
            return 1
        else:
            return 0

    def _optimise_known_target(self):

        """
        Where the target of an invocation is known, provide information about it
        and its context. If a class is being invoked and the conditions are
        appropriate, get information about the specific initialiser.
        """

        if self._should_optimise_known_target() and self._have_known_target():
            last = self.last_op()
            target = last.attr.value
            context = last.attr.context

            # Handle calls to classes.

            if isinstance(target, Class):
                target = target.get_instantiator()
                context = Undefined()

                # A special context is chosen to avoid generating unnecessary
                # context loading and checking instructions.

            return target, context
        else:
            return None

    def _optimise_self_access(self, attrname, classes, node):

        """
        Where the provided 'attrname' accesses an attribute which occupies the
        same position in all possible objects which can be accessed, generate an
        instruction using one of the given 'classes', accessing the attribute
        directly.
        """

        AddressInstruction, AddressContextInstruction, AttrInstruction = classes

        if self._should_optimise_self_access() and self._have_self_input() and \
            not self.unit.is_relocated(attrname):

            # Either generate an instruction operating on an instance attribute.

            try:
                attr = self.unit.parent.instance_attributes()[attrname]
                self.new_op(AttrInstruction(attr))

            # Or generate an instruction operating on a class attribute.

            except KeyError:
                attr = self.unit.parent.all_attributes()[attrname]
                new_attr = attr.via_instance()

                # Only permit loading (not storing) of class attributes via self.

                if AddressContextInstruction is not None:
                    self.new_op(AddressContextInstruction(new_attr))
                else:
                    raise TranslateError(self.module.full_name(), node,
                        "Storing of class attribute %r via self not permitted." % attrname)

            return 1
        else:
            return 0

    def _optimise_temp_storage(self):

        """
        Where the next operation would involve storing a value into temporary
        storage at 'temp_position', record and remove any simple instruction
        which produced the value to be stored such that instead of subsequently
        accessing the temporary storage, that instruction is substituted.

        If no optimisation can be achieved, a StoreTemp instruction is produced
        and the appropriate LoadTemp instruction is returned.

        Restriction: for use only in situations where the source of the
        temporary data will not be disturbed between its first access and its
        subsequent use.
        """

        if self._should_optimise_temp_storage() and \
            self._have_temp_compatible_access():

            last = self.last_op()
            self.remove_op()
            return last
        else:
            return self.get_temp()

    def _optimise_load_operations(self, instruction):

        """
        Incorporate previous load operations into other operations.
        """

        if self._should_optimise_load_operations() and \
            self._have_simple_input() and \
            self._is_simple_input_user(instruction):

            last = self.last_op()
            self.remove_op()
            instruction.input = last

    def _optimise_unused_results(self):

        "Discard results which will not be used."

        if self._have_input():
            self.remove_op()

    # Visitor methods.

    def default(self, node, *args):
        raise TranslateError(self.module.full_name(), node, "Node class %r is not supported." % node.__class__)

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    # Internal helper methods.

    def _visitAttr(self, node, classes):

        """
        Visit the attribute-related 'node', generating instructions based on the
        given 'classes'.
        """

        self.dispatch(node.expr)
        self._generateAttr(node, node.attrname, classes)

    def _generateAttr(self, node, attrname, classes):

        """
        Generate code for the access to 'attrname' using the given 'classes'.
        """

        AddressInstruction, AddressContextInstruction, AttrInstruction, AttrIndexInstruction = classes

        # Where the last operation (defining the attribute owner) yields a
        # constant...

        if self._have_constant_input():
            last = self.active

            # Get the details of the access.

            if isinstance(last.attr, Const):
                target_name = last.attr.value_type_name()
            else:
                target = last.attr.value

                if isinstance(target, Const):
                    target_name = target.value_type_name()
                elif isinstance(target, Instance):
                    target_name = None # skip production of optimised code
                else:
                    target_name = target.full_name()

            # Only try and discover the position if the target can be resolved.

            if target_name is not None:

                # Access the object table to get the attribute position.

                try:
                    table_entry = self.objtable.table[target_name]
                except KeyError:
                    raise TranslateError(self.module.full_name(), node,
                        "No object entry exists for target %r." % target_name)

                try:
                    pos = table_entry[attrname]
                except KeyError:
                    raise TranslateError(self.module.full_name(), node,
                        "No attribute entry exists for name %r in target %r." % (attrname, target_name))

                # Produce a suitable instruction.

                if AddressInstruction is not None:
                    self.replace_op(AddressInstruction(pos))
                else:
                    raise TranslateError(self.module.full_name(), node,
                        "Storing of class or module attribute %r via an object is not permitted." % attrname)

                return

        # Where the last operation involves the special 'self' name, check to
        # see if the attribute is acceptably positioned and produce a direct
        # access to the attribute.

        elif self._optimise_self_access(attrname, (AddressInstruction, AddressContextInstruction, AttrInstruction), node):
            return

        # Otherwise, perform a normal operation.

        try:
            index = self.objtable.get_index(attrname)
        except self.objtable.TableError:
            raise TranslateError(self.module.full_name(), node,
                "No attribute entry exists for name %r." % attrname)

        self.new_op(AttrIndexInstruction(index))

    # Invocations involve the following:
    #
    # 1. Reservation of a frame for the arguments
    # 2. Identification of the target which is then held in temporary storage
    # 3. Optional inclusion of a context (important for methods)
    # 4. Preparation of the argument frame
    # 5. Invocation of the target
    # 6. Discarding of the frame
    #
    # In order to support nested invocations - eg. a(b(c)) - use of the
    # temporary storage is essential.

    def _startCallFunc(self):

        "Record the location of the invocation."

        self.new_op(MakeFrame()) # records the start of the frame

    def _generateCallFunc(self, args, node):

        """
        Support a generic function invocation using the given 'args', occurring
        on the given 'node', where the expression providing the invocation
        target has just been generated.

        In other situations, the invocation is much simpler and does not need to
        handle the full flexibility of a typical Python invocation. Internal
        invocations, such as those employed by operators and certain
        control-flow mechanisms, use predetermined arguments and arguably do not
        need to support the same things as the more general invocations.
        """

        target, context, temp = self._generateCallFuncContext()
        self._generateCallFuncArgs(target, context, temp, args, node)
        return temp

    def _generateCallFuncContext(self):

        """
        Produce code which loads and checks the context of the current
        invocation, the instructions for whose target have already been
        produced, returning a list of instructions which reference the
        invocation target.
        """

        t = self._optimise_known_target()
        if t:
            target, context = t
        else:
            target, context = None, None

        # Store the target in temporary storage for subsequent referencing.

        temp = self._optimise_temp_storage()

        # Where a target or context are not known or where an instance is known
        # to be the context, load the context.

        if context is None or isinstance(context, Instance):
            self.new_op(temp)
            self.new_op(LoadContext())
            self.new_op(StoreFrame(0))

        # Otherwise omit the context.

        else:
            pass # NOTE: Class methods should be supported.

        return target, context, temp

    def _generateCallFuncArgs(self, target, context, temp, args, node):

        """
        Given invocation 'target' and 'context' information, the 'temp'
        reference to the target, a list of nodes representing the 'args'
        (arguments), generate instructions which load the arguments for the
        invocation defined by the given 'node'.
        """

        # Evaluate the arguments.

        employed_positions = set()
        extra_keywords = []

        # Note the presence of the context in the frame where appropriate.

        if context is None or isinstance(context, Instance):
            ncontext = 1
            expect_context = 0
        elif isinstance(context, Class):
            ncontext = 0
            expect_context = 1
        else:
            ncontext = 0
            expect_context = 0

        first = 1
        frame_pos = ncontext

        for arg in args:

            # Handle positional and keyword arguments separately.

            if isinstance(arg, compiler.ast.Keyword):

                # Optimise where the target is known now.

                if target is not None:

                    # Find the parameter table entry for the target.

                    target_name = target.full_name()

                    # Look for a callable with the precise target name.

                    table_entry = self.paramtable.table[target_name]

                    # Look the name up in the parameter table entry.

                    try:
                        pos = table_entry[arg.name]

                    # Where no position is found, this could be an extra keyword
                    # argument.

                    except KeyError:
                        extra_keywords.append(arg)
                        continue

                    # Test for illegal conditions.

                    if pos in employed_positions:
                        raise TranslateError(self.module.full_name(), node,
                            "Keyword argument %r overwrites parameter %r." % (arg.name, pos))

                    employed_positions.add(pos)

                    # Generate code for the keyword and the positioning
                    # operation.

                    self.dispatch(arg.expr)
                    self.new_op(StoreFrame(pos))

                # Otherwise, generate the code needed to obtain the details of
                # the parameter location.

                else:

                    # Combine the target details with the name to get the location.
                    # See the access method on the List class.

                    try:
                        paramindex = self.paramtable.get_index(arg.name)

                    # Where no position is found, this could be an extra keyword
                    # argument.

                    except self.paramtable.TableError:
                        extra_keywords.append(arg)
                        continue

                    # Generate code for the keyword and the positioning
                    # operation.

                    self.dispatch(arg.expr)
                    self.new_op(StoreFrameIndex(paramindex))

                    # use (callable+0)+paramindex+table
                    # checks embedded offset against (callable+0)
                    # moves the current value to frame+position

            else:
                self.dispatch(arg)
                self.new_op(StoreFrame(frame_pos))

                employed_positions.add(frame_pos)

            # Check to see if the first argument is appropriate (compatible with
            # the target where methods are being invoked via classes).

            if first and expect_context:
                continue_label = self.new_label()
                self.new_op(CheckSelf())
                self.new_op(JumpIfTrue(continue_label))

                # Where the context is inappropriate, drop the incomplete frame and
                # raise an exception.

                self.new_op(DropFrame())

                self.new_op(LoadResult())

                self.load_builtin("TypeError", node)
                self.new_op(StoreException())
                self.new_op(RaiseException())
                self.set_label(continue_label)

            first = 0
            frame_pos += 1

        # NOTE: Extra keywords are not supported.
        # NOTE: Somehow, the above needs to be combined with * arguments.

        # Either test for a complete set of arguments.

        if target is not None:

            # Make sure that enough arguments have been given.

            nargs_max = len(target.positional_names)
            ndefaults = len(target.defaults)
            nargs_min = nargs_max - ndefaults

            for i in range(ncontext, nargs_min):
                if i not in employed_positions:
                    raise TranslateError(self.module.full_name(), node,
                        "Argument %r not supplied for %r: need at least %d arguments." % (i+1, target.name, nargs_min))

            nargs = len(args)

            if nargs > nargs_max and not target.has_star and not target.has_dstar:
                raise TranslateError(self.module.full_name(), node,
                    "Too many arguments for %r: need at most %d arguments." % (target.name, nargs_max))

            # Where defaults are involved, put them into the frame.

            self._generateCallFuncDefaultArgs(target, temp, nargs_min, nargs_max, employed_positions)

        # Or generate instructions to do this at run-time.
        # NOTE: CheckFrame has to check the number of arguments and to fill in
        # NOTE: defaults; it also has to shift the invocation frame according to
        # NOTE: the context in use.

        else:
            self.new_op(CheckFrame())

    def _generateCallFuncDefaultArgs(self, target, temp, nargs_min, nargs_max, employed_positions):

        """
        For the given 'target' and 'temp' reference to the target, generate
        default arguments for those positions in the range 'nargs_min'...
        'nargs_max' which are not present in the 'employed_positions'
        collection.
        """

        # Where a lambda is involved, construct a dynamic object to hold the
        # defaults.

        dynamic = target.name is None

        # Here, we use negative index values to visit the right hand end of
        # the defaults list.

        for pos in range(nargs_min, nargs_max):
            if pos not in employed_positions:
                if dynamic:
                    self.new_op(temp)
                    self.new_op(LoadAttr(target.default_attrs[pos - nargs_min]))
                else:
                    self.new_op(LoadAddress(target.default_attrs[pos - nargs_min]))
                self.new_op(StoreFrame(pos))

    def _doCallFunc(self, instruction):

        "Make the invocation."

        self.new_op(instruction)
        self.new_op(JumpWithFrame())

    def _endCallFunc(self, instruction=None):

        "Finish the invocation and tidy up afterwards."

        self.new_op(DropFrame())
        self.new_op(LoadResult())

        # Discard any temporary storage instructions.

        if instruction is not None:
            self.discard_temp(instruction)

    def _generateFunctionDefaults(self, function):

        """
        Generate the default initialisation code for 'function', returning
        a temporary storage reference if a dynamic object was created for the
        function.
        """

        attr_to_default = zip(function.default_attrs, function.defaults)
        if not attr_to_default:
            return None

        # Where a lambda is involved, construct a dynamic object to hold the
        # defaults.

        dynamic = function.name is None

        if dynamic:
            self.new_op(MakeObject(len(attr_to_default)))
            temp = self.get_temp()

        for attr, default in attr_to_default:
            self.dispatch(default)

            self.record_value()
            if dynamic:
                self.new_op(temp)
                self.new_op(StoreAttr(attr))
            else:
                self.new_op(StoreAddress(attr))
            self.set_source()
            self.discard_value()

        if dynamic:
            return temp
        else:
            return None

    def _visitName(self, node, classes):

        """
        Visit the name-related 'node', generating instructions based on the
        given 'classes'.
        """

        name = node.name
        scope = self.get_scope(name)
        #print self.module.name, node.lineno, name, scope
        self._generateName(name, scope, classes, node)

    def _generateName(self, name, scope, classes, node):

        """
        Generate code for the access to 'name' in 'scope' using the given
        'classes', and using the given 'node' as the source of the access.
        """

        NameInstruction, AddressInstruction = classes

        if scope == "local":
            unit = self.unit
            if isinstance(unit, Function):
                self.new_op(NameInstruction(unit.all_locals()[name]))
            elif isinstance(unit, Class):
                self.new_op(AddressInstruction(unit.all_class_attributes()[name]))
            elif isinstance(unit, Module):
                self.new_op(AddressInstruction(unit.module_attributes()[name]))
            else:
                raise TranslateError(self.module.full_name(), node, "Program unit %r has no local %r." % (unit, name))

        elif scope == "global":
            globals = self.module.module_attributes()
            if globals.has_key(name):
                self.new_op(AddressInstruction(globals[name]))
            else:
                raise TranslateError(self.module.full_name(), node, "Module %r has no attribute %r." % (self.module, name))

        else:
            self.new_op(AddressInstruction(self.get_builtin(name, node)))

    def _visitUnary(self, node, method):

        """
        _t = node.expr
        try:
            _result = _t.__pos__()
        except AttributeError:
            raise TypeError
        """

        end_call_label = self.new_label()
        end_label = self.new_label()

        # Evaluate and store the operand in temporary storage.

        self.dispatch(node.expr)
        temp = self._optimise_temp_storage()

        # Produce the invocation.

        self._startCallFunc()
        self.new_op(temp)

        # Get the method on temp.

        self._generateAttr(node, method, self.attribute_load_instructions)

        # Add exception handling to the method acquisition instructions where
        # the attribute access cannot be resolved at compile-time.

        if not self._optimise_known_target():
            self.load_builtin("AttributeError", node)
            self.new_op(CheckException())
            self.new_op(JumpIfTrue(end_call_label))

        temp_method = self._optimise_temp_storage()

        # Add arguments.
        # NOTE: No support for defaults.

        self.new_op(temp) # Explicit context as first argument.
        self._doCallFunc(temp_method)
        self._endCallFunc(temp_method)
        self.new_op(Jump(end_label))

        # End method attempt.

        self.set_label(end_call_label)
        self._endCallFunc() # From the method call.

        # Raise a TypeError.

        self.load_builtin("TypeError", node)
        self.new_op(StoreException())
        self.new_op(RaiseException())

        self.set_label(end_label)

        # Compilation duties...

        self.discard_temp(temp)

    def _visitBinary(self, node, left_method, right_method):

        """
        _t1 = node.left
        _t2 = node.right
        try:
            _result = _t1.__add__(_t2)
            if _result is NotImplemented:
                raise AttributeError
        except AttributeError:
            try:
                _result = _t2.__radd__(_t1)
                if _result is NotImplemented:
                    raise AttributeError
            except AttributeError:
                raise TypeError
        """

        end_left_label = self.new_label()
        right_label = self.new_label()
        end_right_label = self.new_label()
        type_error_label = self.new_label()
        end_label = self.new_label()

        # Evaluate and store the left operand in temporary storage.

        self.dispatch(node.left)
        temp1 = self._optimise_temp_storage()

        # Evaluate and store the right operand in temporary storage.

        self.dispatch(node.right)
        temp2 = self._optimise_temp_storage()

        # Left method.

        self._startCallFunc()
        self.new_op(temp1)

        # Get left method on temp1.

        self._generateAttr(node, left_method, self.attribute_load_instructions)

        # Add exception handling to the method acquisition instructions where
        # the attribute access cannot be resolved at compile-time.

        if not self._optimise_known_target():
            self.load_builtin("AttributeError", node)
            self.new_op(CheckException())
            self.new_op(JumpIfTrue(end_left_label))

        temp_method = self._optimise_temp_storage()

        # Add arguments.
        # NOTE: No support for defaults.

        self.new_op(temp1) # Explicit context as first argument.
        self.new_op(temp2)
        self._doCallFunc(temp_method)
        self._endCallFunc(temp_method)

        # Test for NotImplemented.
        # Don't actually raise an exception.

        self.new_op(TestIdentityAddress(self.get_builtin("NotImplemented", node)))
        self.new_op(JumpIfTrue(right_label))
        self.new_op(Jump(end_label))

        # End left method attempt.

        self.set_label(end_left_label)
        self._endCallFunc() # From the left method call.

        # Right method.

        self.set_label(right_label)
        self._startCallFunc()
        self.new_op(temp2)

        # Get right method on temp2.

        self._generateAttr(node, right_method, self.attribute_load_instructions)

        # Add exception handling to the method acquisition instructions where
        # the attribute access cannot be resolved at compile-time.

        if not self._optimise_known_target():
            self.load_builtin("AttributeError", node)
            self.new_op(CheckException())
            self.new_op(JumpIfTrue(end_right_label))

        temp_method = self._optimise_temp_storage()

        # Add arguments.
        # NOTE: No support for defaults.

        self.new_op(temp2) # Explicit context as first argument.
        self.new_op(temp1)
        self._doCallFunc(temp_method)
        self._endCallFunc(temp_method)

        # Test for NotImplemented.
        # Don't actually raise an exception.

        self.new_op(TestIdentityAddress(self.get_builtin("NotImplemented", node)))
        self.new_op(JumpIfTrue(type_error_label))
        self.new_op(Jump(end_label))

        # End right method attempt.

        self.set_label(end_right_label)
        self._endCallFunc() # From the right method call.

        # Raise a TypeError.

        self.set_label(type_error_label)
        self.load_builtin("TypeError", node)
        self.new_op(StoreException())
        self.new_op(RaiseException())

        self.set_label(end_label)

        # Compilation duties...

        self.discard_temp(temp1)
        self.discard_temp(temp2)

    # Concrete visitor methods.

    def visitAdd(self, node):
        self._visitBinary(node, "__add__", "__radd__")

    def visitAnd(self, node):
        next_label = self.new_label()

        for n in node.nodes[:-1]:
            self.dispatch(n)
            self.new_op(TestBoolean())
            self.new_op(JumpIfFalse(next_label))

        self.dispatch(node.nodes[-1])
        self.set_label(next_label)

        # Prevent incorrect optimisation.

        self.active = None

    def visitAssert(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Assert")

    def visitAssign(self, node):
        self.dispatch(node.expr)
        self.record_value()

        for n in node.nodes:
            self.dispatch(n)

        self.discard_value()

    def visitAssAttr(self, node):
        self._visitAttr(node, self.attribute_store_instructions)
        self.set_source()

    def visitAssList(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "AssList")

    def visitAssName(self, node):
        self._visitName(node, self.name_store_instructions)
        self.set_source()

    visitAssTuple = visitAssList

    def visitAugAssign(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "AugAssign")

    def visitBackquote(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Backquote")

    def visitBitand(self, node):
        self._visitBinary(node, "__and__", "__rand__")

    def visitBitor(self, node):
        self._visitBinary(node, "__or__", "__ror__")

    def visitBitxor(self, node):
        self._visitBinary(node, "__xor__", "__rxor__")

    def visitBreak(self, node):
        next_label, exit_label = self.get_loop_labels()
        self.new_op(Jump(exit_label))

    def visitCallFunc(self, node):

        """
        Evaluate positional arguments, evaluate and store keyword arguments in
        the correct location, then invoke the function.
        """

        # Mark the frame, evaluate the target, generate the call.

        self._startCallFunc()
        self.dispatch(node.node)
        temp = self._generateCallFunc(node.args, node)
        self._doCallFunc(temp)
        self._endCallFunc(temp)

    def visitClass(self, node):

        # Store the name.

        self.new_op(LoadConst(node.unit))
        self.record_value()
        self._visitName(node, self.name_store_instructions)
        self.set_source()
        self.discard_value()

        # Visit the code.

        unit = self.unit
        self.unit = node.unit
        self.unit.code_location = self.module.code_location # class body code is not independently addressable
        self.dispatch(node.code)
        self.unit = unit

    def visitCompare(self, node):

        """
        self.dispatch(node.expr)
        for op_name, next_node in compare.ops:
            methods = self.comparison_methods[op_name]
            if methods is not None:
                # Generate method call using evaluated argument and next node.
            else:
                # Deal with the special operators.
            # Provide short-circuiting.
        """

    def visitConst(self, node):
        const = self.module.constant_values[node.value]
        self.new_op(LoadConst(const))

    def visitContinue(self, node):
        next_label, exit_label = self.get_loop_labels()
        self.new_op(Jump(next_label))

    def visitDecorators(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Decorators")

    def visitDict(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Dict")

    def visitDiscard(self, node):
        self.dispatch(node.expr)
        self._optimise_unused_results()

    def visitDiv(self, node):
        self._visitBinary(node, "__div__", "__rdiv__")

    def visitEllipsis(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Ellipsis")

    def visitExec(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Exec")

    def visitExpression(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Expression")

    def visitFloorDiv(self, node):
        self._visitBinary(node, "__floordiv__", "__rfloordiv__")

    def visitFor(self, node):
        exit_label = self.new_label()
        next_label = self.new_label()
        else_label = self.new_label()

        # Get the "list" to be iterated over, obtain its iterator.

        self._startCallFunc()
        self.dispatch(node.list)
        self._generateAttr(node, "__iter__", self.attribute_load_instructions)
        temp = self._generateCallFunc([], node)
        self._doCallFunc(temp)
        self._endCallFunc(temp)

        temp_iterator = self._optimise_temp_storage()

        # In the loop...

        self.set_label(next_label)

        # Use the iterator to get the next value.

        self._startCallFunc()
        self.new_op(temp_iterator)
        self._generateAttr(node, "next", self.attribute_load_instructions)
        temp = self._generateCallFunc([], node)
        self._doCallFunc(temp)
        self._endCallFunc(temp)

        # Test for StopIteration.

        self.load_builtin("StopIteration", node)
        self.new_op(CheckException())
        if node.else_ is not None:
            self.new_op(JumpIfTrue(else_label))
        else:
            self.new_op(JumpIfTrue(exit_label))

        # Assign to the target.

        self.dispatch(node.assign)

        # Process the body with the current next and exit points.

        self.add_loop_labels(next_label, exit_label)
        self.dispatch(node.body)
        self.drop_loop_labels()

        # Repeat the loop.

        self.new_op(Jump(next_label))

        # Produce the "else" section.

        if node.else_ is not None:
            self.set_label(exit_label)
            self.dispatch(node.else_)

        # After the loop...

        self.set_label(exit_label)

        # Compilation duties...

        self.discard_temp(temp_iterator)

    def visitFrom(self, node): pass

    def visitFunction(self, node):

        # Only store the name when visiting this node from outside.

        if self.unit is not node.unit:
            self.new_op(LoadConst(node.unit))

            self.record_value()
            self._visitName(node, self.name_store_instructions) # AssName equivalent
            self.set_source()
            self.discard_value()

            self._generateFunctionDefaults(node.unit)

        # Visiting of the code occurs when get_code is invoked on this node.

        else:
            self.dispatch(node.code)
            if not isinstance(self.last_op(), Return):
                self.dispatch(compiler.ast.Name("None"))
                self.new_op(StoreResult())

                self.new_op(Return())

    def visitGenExpr(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "GenExpr")

    def visitGenExprFor(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "GenExprFor")

    def visitGenExprIf(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "GenExprIf")

    def visitGenExprInner(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "GenExprInner")

    def visitGetattr(self, node):
        self._visitAttr(node, self.attribute_load_instructions)

    def visitGlobal(self, node): pass

    def visitIf(self, node):
        first = 1
        exit_label = self.new_label()

        clauses = node.tests + [(None, node.else_)]
        last_clause = clauses[-1]

        for clause in clauses:
            test, body = clause
            if body is None:
                break
            if not first:
                self.set_label(next_label)
            if test is not None:
                self.dispatch(test)
                next_label = self.new_label()
                self.new_op(JumpIfFalse(next_label))
            self.dispatch(body)
            if clause is not last_clause:
                self.new_op(Jump(exit_label))
            first = 0

        self.set_label(exit_label)

    def visitImport(self, node): pass

    def visitInvert(self, node):
        self._visitUnary(node, "__invert__")

    def visitKeyword(self, node): pass

    def visitLambda(self, node):

        # Produce the reference to this function when visiting this node from
        # outside.

        if self.unit is not node.unit:
            temp = self._generateFunctionDefaults(node.unit)
            self.new_op(LoadConst(node.unit))

            # Populate the new object required for the function.

            if temp is not None:
                self.record_value()
                self.new_op(temp)
                self.new_op(StoreCallable())
                self.set_source()
                self.discard_value()

                self.new_op(temp)
                self.discard_temp(temp)

        # Visiting of the code occurs when get_code is invoked on this node.

        else:
            self.dispatch(node.code)
            self.new_op(StoreResult())
            self.new_op(Return())

    def visitLeftShift(self, node):
        self._visitBinary(node, "__lshift__", "__rlshift__")

    def visitList(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "List")

    def visitListComp(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "ListComp")

    def visitListCompFor(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "ListCompFor")

    def visitListCompIf(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "ListCompIf")

    def visitMod(self, node):
        self._visitBinary(node, "__mod__", "__rmod__")

    def visitModule(self, node):
        self.dispatch(node.node)

    def visitMul(self, node):
        self._visitBinary(node, "__mul__", "__rmul__")

    def visitName(self, node):
        if node.name == "None":
            const = self.module.constant_values[None]
            self.new_op(LoadConst(const))
        else:
            self._visitName(node, self.name_load_instructions)

    def visitNot(self, node):
        next_label = self.new_label()
        true_label = self.new_label()

        self.dispatch(node.expr)
        self.new_op(TestBoolean())
        self.new_op(JumpIfTrue(true_label))
        self.load_builtin("True", node)
        self.new_op(Jump(next_label))

        self.set_label(true_label)
        self.load_builtin("False", node)
        self.set_label(next_label)

        # Prevent incorrect optimisation.

        self.active = None

    def visitOr(self, node):
        next_label = self.new_label()

        for n in node.nodes[:-1]:
            self.dispatch(n)
            self.new_op(TestBoolean())
            self.new_op(JumpIfTrue(next_label))

        self.dispatch(node.nodes[-1])
        self.set_label(next_label)

        # Prevent incorrect optimisation.

        self.active = None

    def visitPass(self, node): pass

    def visitPower(self, node):
        self._visitBinary(node, "__pow__", "__rpow__")

    def visitPrint(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Print")

    def visitPrintnl(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Printnl")

    def visitRaise(self, node):
        # NOTE: expr1 only => instance provided
        self.dispatch(node.expr1)

        if node.expr2 is not None:
            temp = self._optimise_temp_storage()

            self.dispatch(node.expr2)
            temp_arg = self._optimise_temp_storage()

            self._startCallFunc()
            self.new_op(temp_arg)
            self.new_op(StoreFrame(0))
            self._doCallFunc(temp)
            self._endCallFunc(temp)

            self.discard_temp(temp_arg)

        self.new_op(StoreException())
        self.new_op(RaiseException())

    def visitReturn(self, node):
        if node.value is not None:
            self.dispatch(node.value)
        else:
            self.dispatch(compiler.ast.Name("None"))

        self.new_op(StoreResult())
        self.new_op(Return())

    def visitRightShift(self, node):
        self._visitBinary(node, "__rshift__", "__rrshift__")

    def visitSlice(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Slice")

    def visitStmt(self, node):
        for n in node.nodes:
            self.dispatch(n)

    def visitSub(self, node):
        self._visitBinary(node, "__sub__", "__rsub__")

    def visitSubscript(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Subscript")

    def visitTryExcept(self, node):
        exit_label = self.new_label()
        success_label = self.new_label()
        handler_label = self.new_label()

        self.add_exception_labels(handler_label, exit_label)

        # Try...
        # Produce the code, then jump to the exit.

        self.new_op(PushHandler(handler_label))
        self.dispatch(node.body)
        self.new_op(PopHandler())
        self.new_op(Jump(exit_label))

        # Start of handlers.

        self.set_label(handler_label)
        self.new_op(PopHandler())

        for name, assignment, handler in node.handlers:
            next_label = self.new_label()

            # Test the given exception against the current exception.

            if name is not None:
                self.dispatch(name)
                self.new_op(CheckException())
                self.new_op(JumpIfFalse(next_label))

            # Handle assignment to exception variable.

            if assignment is not None:
                self.dispatch(assignment)

            # Produce the handler code, then jump to the exit.

            self.dispatch(handler)
            self.new_op(Jump(exit_label))

            self.set_label(next_label)

        # Unhandled exceptions.

        #self.new_op(LoadException())
        self.new_op(RaiseException())

        # Optional else clause.

        if node.else_ is not None:
            self.dispatch(node.else_)

        self.set_label(exit_label)
        self.drop_exception_labels()

    def visitTryFinally(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "TryFinally")

    def visitTuple(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Tuple")

    def visitUnaryAdd(self, node):
        self._visitUnary(node, "__pos__")

    def visitUnarySub(self, node):
        self._visitUnary(node, "__neg__")

    def visitWhile(self, node):
        exit_label = self.new_label()
        next_label = self.new_label()
        else_label = self.new_label()

        self.set_label(next_label)
        self.dispatch(node.test)
        if node.else_ is not None:
            self.new_op(JumpIfFalse(else_label))
        else:
            self.new_op(JumpIfFalse(exit_label))

        self.add_loop_labels(next_label, exit_label)

        self.dispatch(node.body)
        self.new_op(Jump(next_label))

        if node.else_ is not None:
            self.set_label(else_label)
            self.dispatch(node.else_)

        self.set_label(exit_label)
        self.drop_loop_labels()

        # Prevent incorrect optimisation.

        self.active = None

    def visitWith(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "With")

    def visitYield(self, node): raise TranslationNotImplementedError(self.module.full_name(), node, "Yield")

    # Useful data.

    comparison_methods = {
        "==" : ("__eq__", "__ne__"),
        "!=" : ("__ne__", "__eq__"),
        "<" : ("__lt__", "__gt__"),
        "<=" : ("__le__", "__ge__"),
        ">=" : ("__ge__", "__le__"),
        ">" : ("__gt__", "__lt__"),
        "is" : None,
        "is not" : None,
        "in" : None,
        "not in" : None
        }

# vim: tabstop=4 expandtab shiftwidth=4
